from arsenal.maths.util import *
from arsenal.maths.rvs import *
from arsenal.maths.compare import compare, check_equal, pp_plot
from arsenal.maths import stepsize
from arsenal.maths.checkgrad import fdcheck, quick_fdcheck
from arsenal.maths.optimize import NProx, NProj
